# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import TYPE_CHECKING, List, Union, Optional
from datetime import datetime

import httpx

from ...types import (
    EventIngestResponse,
    EventSearchResponse,
    EventUpdateResponse,
    EventDeprecateResponse,
    event_ingest_params,
    event_search_params,
    event_update_params,
)
from ..._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from ..._utils import maybe_transform
from .backfills import (
    Backfills,
    AsyncBackfills,
    BackfillsWithRawResponse,
    AsyncBackfillsWithRawResponse,
)
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import to_raw_response_wrapper, async_to_raw_response_wrapper
from ..._base_client import make_request_options

if TYPE_CHECKING:
    from ..._client import Orb, AsyncOrb

__all__ = ["Events", "AsyncEvents"]


class Events(SyncAPIResource):
    backfills: Backfills
    with_raw_response: EventsWithRawResponse

    def __init__(self, client: Orb) -> None:
        super().__init__(client)
        self.backfills = Backfills(client)
        self.with_raw_response = EventsWithRawResponse(self)

    def update(
        self,
        event_id: str,
        *,
        event_name: str,
        properties: object,
        timestamp: Union[str, datetime],
        customer_id: Optional[str] | NotGiven = NOT_GIVEN,
        external_customer_id: Optional[str] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> EventUpdateResponse:
        """
        This endpoint is used to amend a single usage event with a given `event_id`.
        `event_id` refers to the `idempotency_key` passed in during ingestion. The event
        will maintain its existing `event_id` after the amendment.

        This endpoint will mark the existing event as ignored, and Orb will only use the
        new event passed in the body of this request as the source of truth for that
        `event_id`. Note that a single event can be amended any number of times, so the
        same event can be overwritten in subsequent calls to this endpoint, or
        overwritten using the [Amend customer usage](amend-usage) endpoint. Only a
        single event with a given `event_id` will be considered the source of truth at
        any given time.

        This is a powerful and audit-safe mechanism to retroactively update a single
        event in cases where you need to:

        - update an event with new metadata as you iterate on your pricing model
        - update an event based on the result of an external API call (e.g. call to a
          payment gateway succeeded or failed)

        This amendment API is always audit-safe. The process will still retain the
        original event, though it will be ignored for billing calculations. For auditing
        and data fidelity purposes, Orb never overwrites or permanently deletes ingested
        usage data.

        ## Request validation

        - The `timestamp` of the new event must match the `timestamp` of the existing
          event already ingested. As with ingestion, all timestamps must be sent in
          ISO8601 format with UTC timezone offset.
        - The `customer_id` or `external_customer_id` of the new event must match the
          `customer_id` or `external_customer_id` of the existing event already
          ingested. Exactly one of `customer_id` and `external_customer_id` should be
          specified, and similar to ingestion, the ID must identify a Customer resource
          within Orb. Unlike ingestion, for event amendment, we strictly enforce that
          the Customer must be in the Orb system, even during the initial integration
          period. We do not allow updating the `Customer` an event is associated with.
        - Orb does not accept an `idempotency_key` with the event in this endpoint,
          since this request is by design idempotent. On retryable errors, you should
          retry the request and assume the amendment operation has not succeeded until
          receipt of a 2xx.
        - The event's `timestamp` must fall within the customer's current subscription's
          billing period, or within the grace period of the customer's current
          subscription's previous billing period.

        Args:
          event_name: A name to meaningfully identify the action or event type.

          properties: A dictionary of custom properties. Values in this dictionary must be numeric,
              boolean, or strings. Nested dictionaries are disallowed.

          timestamp: An ISO 8601 format date with no timezone offset (i.e. UTC). This should
              represent the time that usage was recorded, and is particularly important to
              attribute usage to a given billing period.

          customer_id: The Orb Customer identifier

          external_customer_id: An alias for the Orb customer, whose mapping is specified when creating the
              customer

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._put(
            f"/events/{event_id}",
            body=maybe_transform(
                {
                    "event_name": event_name,
                    "properties": properties,
                    "timestamp": timestamp,
                    "customer_id": customer_id,
                    "external_customer_id": external_customer_id,
                },
                event_update_params.EventUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=EventUpdateResponse,
        )

    def deprecate(
        self,
        event_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> EventDeprecateResponse:
        """
        This endpoint is used to deprecate a single usage event with a given `event_id`.
        `event_id` refers to the `idempotency_key` passed in during ingestion.

        This endpoint will mark the existing event as ignored. Note that if you attempt
        to re-ingest an event with the same `event_id` as a deprecated event, Orb will
        return an error.

        This is a powerful and audit-safe mechanism to retroactively deprecate a single
        event in cases where you need to:

        - no longer bill for an event that was improperly reported
        - no longer bill for an event based on the result of an external API call (e.g.
          call to a payment gateway failed and the user should not be billed)

        If you want to only change specific properties of an event, but keep the event
        as part of the billing calculation, use the [Amend single event](amend-event)
        endpoint instead.

        This API is always audit-safe. The process will still retain the deprecated
        event, though it will be ignored for billing calculations. For auditing and data
        fidelity purposes, Orb never overwrites or permanently deletes ingested usage
        data.

        ## Request validation

        - Orb does not accept an `idempotency_key` with the event in this endpoint,
          since this request is by design idempotent. On retryable errors, you should
          retry the request and assume the deprecation operation has not succeeded until
          receipt of a 2xx.
        - The event's `timestamp` must fall within the customer's current subscription's
          billing period, or within the grace period of the customer's current
          subscription's previous billing period. Orb does not allow deprecating events
          for billing periods that have already invoiced customers.
        - The `customer_id` or the `external_customer_id` of the original event
          ingestion request must identify a Customer resource within Orb, even if this
          event was ingested during the initial integration period. We do not allow
          deprecating events for customers not in the Orb system.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._put(
            f"/events/{event_id}/deprecate",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=EventDeprecateResponse,
        )

    def ingest(
        self,
        *,
        events: List[event_ingest_params.Event],
        backfill_id: Optional[str] | NotGiven = NOT_GIVEN,
        debug: bool | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> EventIngestResponse:
        """Orb's event ingestion model and API is designed around two core principles:

        1.

        **Data fidelity**: The accuracy of your billing model depends on a robust
           foundation of events. Orb's API protocol encourages usage patterns that
           ensure that your data is consistently complete and correct.
        2. **Fast integration**: Sending events into Orb requires no tedious setup steps
           or explicit field schema for your event shape, making it instant to start
           streaming in usage in real-time.

        ## Event shape

        Events are the starting point for all usage calculations in the system, and are
        simple at their core:

        ```ts
        {
          // customer_id and external_customer_id are used to
          // attribute usage to a given Customer. Exactly one of these
          // should be specified in a given ingestion event.

          // `customer_id` is the Orb generated identifier for the Customer,
          // which is returned from the Create customer API call.
          customer_id: string,

          // external_customer_id is an alternate identifier which is associated
          // with a Customer at creation time. This is treated as an alias for
          // customer_id, and is usually set to an identifier native to your system.
          external_customer_id: string,

          // A string name identifying the event, usually a usage
          // action. By convention, this should not contain any whitespace.
          event_name: string,

          // An ISO 8601 format date with no timezone offset.
          // This should represent the time that usage occurred
          // and is important to attribute usage to a given
          // billing period. See the notes below on determining the timestamp.
          // e.g. 2020-12-09T16:09:53Z
          timestamp: string,

          // A unique value, generated by the client, that is
          // used to de-duplicate events.
          // Exactly one event with a given
          // idempotency key will be ingested, which allows for
          // safe request retries.
          idempotency_key: string

          // Optional custom metadata to attach to the event.
          // This might include a numeric value used for aggregation,
          // or a string/boolean value used for filtering.
          // The schema of this dictionary need not be pre-declared, and
          // properties can be added at any time.
          properties: {
            [key: string]?: string | number | boolean,
          },
        }
        ```

        ## Required fields

        Because events streamed to Orb are meant to be as flexible as possible, there
        are only a few required fields in every event.

        - We recommend that `idempotency_key` are unique strings that you generated with
          V4 UUIDs, but only require that they uniquely identify an event (i.e. don’t
          collide).
        - The `timestamp` field in the event body will be used to determine which
          billable period a given event falls into. For example, with a monthly billing
          cycle starting from the first of December, Orb will calculate metrics based on
          events that fall into the range
          `12-01 00:00:00 <= timestamp < 01-01 00:00:00`.

        ## Logging metadata

        Orb allows tagging events with metadata using a flexible properties dictionary.
        Since Orb does not enforce a rigid schema for this field-set, key-value pairs
        can be added dynamically as your events evolve.

        This dictionary can be helpful for a wide variety of use cases:

        - Numeric properties on events like `compute_time_ms` can later be inputs to our
          flexible query engine to determine usage.
        - Logging a region or cluster with each event can help you provide customers
          more granular visibility into their usage.

        We encourage logging this metadata with an eye towards future use cases to
        ensure full coverage for historical data. The datatype of the value in the
        properties dictionary is important for metric creation from an event source.
        Values that you wish to numerically aggregate should be of numeric type in the
        event.

        ## Determining event timestamp

        For cases where usage is being reported in real time as it is occurring,
        timestamp should correspond to the time that usage occurred.

        In cases where usage is reported in aggregate for a historical timeframe at a
        regular interval, we recommend setting the event `timestamp` to the midpoint of
        the interval. As an example, if you have an hourly reporter that sends data once
        an hour for the previous hour of usage, setting the `timestamp` to the half-hour
        mark will ensure that the usage is counted within the correct period.

        Note that other time-related fields (e.g. time elapsed) can be added to the
        properties dictionary as necessary.

        In cases where usage is reported in aggregate for a historical timeframe, the
        timestamp must be within the grace period set for your account. Events with
        `timestamp < current_time - grace_period` will not be accepted as a valid event,
        and will throw validation errors. Enforcing the grace period enables Orb to
        accurately map usage to the correct billing cycle and ensure that all usage is
        billed for in the corresponding billing period.

        ## Event validation

        Orb’s validation ensures that you recognize errors in your events as quickly as
        possible, and the API provides informative error messages to help you fix
        problems quickly.

        We validate the following:

        - Exactly one of `customer_id` and `external_customer_id` should be specified.
        - If specified, `customer_id` must identify a Customer resource within Orb. We
          do not support sending events for customers that have not been provisioned.
          Similarly, if specified, `external_customer_id` must be an identifier that is
          associated with an Orb Customer resource. Note: During our initial integration
          period, this enforcement will be temporarily turned into a warning to ensure
          smooth customer migration.
        - `timestamp` must conform to ISO 8601 and represent a timestamp at most 1 hour
          in the future. This timestamp should be sent in UTC timezone (no timezone
          offset).

        ## Idempotency and retry semantics

        Orb's idempotency guarantees allow you to implement safe retry logic in the
        event of network or machine failures, ensuring data fidelity. Each event in the
        request payload is associated with an idempotency key, and Orb guarantees that a
        single idempotency key will be successfully ingested at most once.

        - Successful responses return a 200 HTTP status code. The response contains
          information about previously processed events.
        - Requests that return a `4xx` HTTP status code indicate a payload error and
          contain at least one event with a validation failure. An event with a
          validation failure can be re-sent to the ingestion endpoint (after the payload
          is fixed) with the original idempotency key since that key is not marked as
          processed.
        - Requests that return a `5xx` HTTP status code indicate a server-side failure.
          These requests should be retried in their entirety.

        ## API usage and limits

        The ingestion API is designed made for real-time streaming ingestion and
        architected for high throughput. Even if events are later deemed unnecessary or
        filtered out, we encourage you to log them to Orb if they may be relevant to
        billing calculations in the future.

        To take advantage of the real-time features of the Orb platform and avoid any
        chance of dropped events by producers, we recommend reporting events to Orb
        frequently. Optionally, events can also be briefly aggregated at the source, as
        this API accepts an array of event bodies.

        Orb does not currently enforce a hard rate-limit for API usage or a maximum
        request payload size, but please give us a heads up if you’re changing either of
        these factors by an order of magnitude from initial setup.

        ## Testing in debug mode

        The ingestion API supports a debug mode, which returns additional verbose output
        to indicate which event idempotency keys were newly ingested or duplicates from
        previous requests. To enable this mode, mark `debug=true` as a query parameter.

        If `debug=true` is not specified, the response will only contain
        `validation_failed`. Orb will still honor the idempotency guarantees set
        [here](../guides/events-and-metrics/event-ingestion#event-volume-and-concurrency)
        in all cases.

        We strongly recommend that you only use debug mode as part of testing your
        initial Orb integration. Once you're ready to switch to production, disable
        debug mode to take advantage of improved performance and maximal throughput.

        #### Example: ingestion response with `debug=true`

        ```json
        {
          "debug": {
            "duplicate": [],
            "ingested": ["B7E83HDMfJPAunXW", "SJs5DQJ3TnwSqEZE", "8SivfDsNKwCeAXim"]
          },
          "validation_failed": []
        }
        ```

        #### Example: ingestion response with `debug=false`

        ```json
        {
          "validation_failed": []
        }
        ```

        Args:
          backfill_id: If this ingestion request is part of a backfill, this parameter ties the
              ingested events to the backfill

          debug: Flag to enable additional debug information in the endpoint response

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/ingest",
            body=maybe_transform({"events": events}, event_ingest_params.EventIngestParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
                query=maybe_transform(
                    {
                        "backfill_id": backfill_id,
                        "debug": debug,
                    },
                    event_ingest_params.EventIngestParams,
                ),
            ),
            cast_to=EventIngestResponse,
        )

    def search(
        self,
        *,
        event_ids: List[str],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> EventSearchResponse:
        """
        This endpoint returns a filtered set of events for an account in a
        [paginated list format](../reference/pagination).

        Note that this is a `POST` endpoint rather than a `GET` endpoint because it
        employs a JSON body for search criteria rather than query parameters, allowing
        for a more flexible search syntax.

        Note that a search criteria _must_ be specified. Currently, Orb supports the
        following criteria:

        - `event_ids`: This is an explicit array of IDs to filter by. Note that an
          event's ID is the `idempotency_key` that was originally used for ingestion.

        By default, Orb does not return _deprecated_ events in this endpoint.

        By default, Orb will not throw a `404` if no events matched, Orb will return an
        empty array for `data` instead.

        Args:
          event_ids: This is an explicit array of IDs to filter by. Note that an event's ID is the
              idempotency_key that was originally used for ingestion. Values in this array
              will be treated case sensitively.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/events/search",
            body=maybe_transform({"event_ids": event_ids}, event_search_params.EventSearchParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=EventSearchResponse,
        )


class AsyncEvents(AsyncAPIResource):
    backfills: AsyncBackfills
    with_raw_response: AsyncEventsWithRawResponse

    def __init__(self, client: AsyncOrb) -> None:
        super().__init__(client)
        self.backfills = AsyncBackfills(client)
        self.with_raw_response = AsyncEventsWithRawResponse(self)

    async def update(
        self,
        event_id: str,
        *,
        event_name: str,
        properties: object,
        timestamp: Union[str, datetime],
        customer_id: Optional[str] | NotGiven = NOT_GIVEN,
        external_customer_id: Optional[str] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> EventUpdateResponse:
        """
        This endpoint is used to amend a single usage event with a given `event_id`.
        `event_id` refers to the `idempotency_key` passed in during ingestion. The event
        will maintain its existing `event_id` after the amendment.

        This endpoint will mark the existing event as ignored, and Orb will only use the
        new event passed in the body of this request as the source of truth for that
        `event_id`. Note that a single event can be amended any number of times, so the
        same event can be overwritten in subsequent calls to this endpoint, or
        overwritten using the [Amend customer usage](amend-usage) endpoint. Only a
        single event with a given `event_id` will be considered the source of truth at
        any given time.

        This is a powerful and audit-safe mechanism to retroactively update a single
        event in cases where you need to:

        - update an event with new metadata as you iterate on your pricing model
        - update an event based on the result of an external API call (e.g. call to a
          payment gateway succeeded or failed)

        This amendment API is always audit-safe. The process will still retain the
        original event, though it will be ignored for billing calculations. For auditing
        and data fidelity purposes, Orb never overwrites or permanently deletes ingested
        usage data.

        ## Request validation

        - The `timestamp` of the new event must match the `timestamp` of the existing
          event already ingested. As with ingestion, all timestamps must be sent in
          ISO8601 format with UTC timezone offset.
        - The `customer_id` or `external_customer_id` of the new event must match the
          `customer_id` or `external_customer_id` of the existing event already
          ingested. Exactly one of `customer_id` and `external_customer_id` should be
          specified, and similar to ingestion, the ID must identify a Customer resource
          within Orb. Unlike ingestion, for event amendment, we strictly enforce that
          the Customer must be in the Orb system, even during the initial integration
          period. We do not allow updating the `Customer` an event is associated with.
        - Orb does not accept an `idempotency_key` with the event in this endpoint,
          since this request is by design idempotent. On retryable errors, you should
          retry the request and assume the amendment operation has not succeeded until
          receipt of a 2xx.
        - The event's `timestamp` must fall within the customer's current subscription's
          billing period, or within the grace period of the customer's current
          subscription's previous billing period.

        Args:
          event_name: A name to meaningfully identify the action or event type.

          properties: A dictionary of custom properties. Values in this dictionary must be numeric,
              boolean, or strings. Nested dictionaries are disallowed.

          timestamp: An ISO 8601 format date with no timezone offset (i.e. UTC). This should
              represent the time that usage was recorded, and is particularly important to
              attribute usage to a given billing period.

          customer_id: The Orb Customer identifier

          external_customer_id: An alias for the Orb customer, whose mapping is specified when creating the
              customer

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._put(
            f"/events/{event_id}",
            body=maybe_transform(
                {
                    "event_name": event_name,
                    "properties": properties,
                    "timestamp": timestamp,
                    "customer_id": customer_id,
                    "external_customer_id": external_customer_id,
                },
                event_update_params.EventUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=EventUpdateResponse,
        )

    async def deprecate(
        self,
        event_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> EventDeprecateResponse:
        """
        This endpoint is used to deprecate a single usage event with a given `event_id`.
        `event_id` refers to the `idempotency_key` passed in during ingestion.

        This endpoint will mark the existing event as ignored. Note that if you attempt
        to re-ingest an event with the same `event_id` as a deprecated event, Orb will
        return an error.

        This is a powerful and audit-safe mechanism to retroactively deprecate a single
        event in cases where you need to:

        - no longer bill for an event that was improperly reported
        - no longer bill for an event based on the result of an external API call (e.g.
          call to a payment gateway failed and the user should not be billed)

        If you want to only change specific properties of an event, but keep the event
        as part of the billing calculation, use the [Amend single event](amend-event)
        endpoint instead.

        This API is always audit-safe. The process will still retain the deprecated
        event, though it will be ignored for billing calculations. For auditing and data
        fidelity purposes, Orb never overwrites or permanently deletes ingested usage
        data.

        ## Request validation

        - Orb does not accept an `idempotency_key` with the event in this endpoint,
          since this request is by design idempotent. On retryable errors, you should
          retry the request and assume the deprecation operation has not succeeded until
          receipt of a 2xx.
        - The event's `timestamp` must fall within the customer's current subscription's
          billing period, or within the grace period of the customer's current
          subscription's previous billing period. Orb does not allow deprecating events
          for billing periods that have already invoiced customers.
        - The `customer_id` or the `external_customer_id` of the original event
          ingestion request must identify a Customer resource within Orb, even if this
          event was ingested during the initial integration period. We do not allow
          deprecating events for customers not in the Orb system.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._put(
            f"/events/{event_id}/deprecate",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=EventDeprecateResponse,
        )

    async def ingest(
        self,
        *,
        events: List[event_ingest_params.Event],
        backfill_id: Optional[str] | NotGiven = NOT_GIVEN,
        debug: bool | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> EventIngestResponse:
        """Orb's event ingestion model and API is designed around two core principles:

        1.

        **Data fidelity**: The accuracy of your billing model depends on a robust
           foundation of events. Orb's API protocol encourages usage patterns that
           ensure that your data is consistently complete and correct.
        2. **Fast integration**: Sending events into Orb requires no tedious setup steps
           or explicit field schema for your event shape, making it instant to start
           streaming in usage in real-time.

        ## Event shape

        Events are the starting point for all usage calculations in the system, and are
        simple at their core:

        ```ts
        {
          // customer_id and external_customer_id are used to
          // attribute usage to a given Customer. Exactly one of these
          // should be specified in a given ingestion event.

          // `customer_id` is the Orb generated identifier for the Customer,
          // which is returned from the Create customer API call.
          customer_id: string,

          // external_customer_id is an alternate identifier which is associated
          // with a Customer at creation time. This is treated as an alias for
          // customer_id, and is usually set to an identifier native to your system.
          external_customer_id: string,

          // A string name identifying the event, usually a usage
          // action. By convention, this should not contain any whitespace.
          event_name: string,

          // An ISO 8601 format date with no timezone offset.
          // This should represent the time that usage occurred
          // and is important to attribute usage to a given
          // billing period. See the notes below on determining the timestamp.
          // e.g. 2020-12-09T16:09:53Z
          timestamp: string,

          // A unique value, generated by the client, that is
          // used to de-duplicate events.
          // Exactly one event with a given
          // idempotency key will be ingested, which allows for
          // safe request retries.
          idempotency_key: string

          // Optional custom metadata to attach to the event.
          // This might include a numeric value used for aggregation,
          // or a string/boolean value used for filtering.
          // The schema of this dictionary need not be pre-declared, and
          // properties can be added at any time.
          properties: {
            [key: string]?: string | number | boolean,
          },
        }
        ```

        ## Required fields

        Because events streamed to Orb are meant to be as flexible as possible, there
        are only a few required fields in every event.

        - We recommend that `idempotency_key` are unique strings that you generated with
          V4 UUIDs, but only require that they uniquely identify an event (i.e. don’t
          collide).
        - The `timestamp` field in the event body will be used to determine which
          billable period a given event falls into. For example, with a monthly billing
          cycle starting from the first of December, Orb will calculate metrics based on
          events that fall into the range
          `12-01 00:00:00 <= timestamp < 01-01 00:00:00`.

        ## Logging metadata

        Orb allows tagging events with metadata using a flexible properties dictionary.
        Since Orb does not enforce a rigid schema for this field-set, key-value pairs
        can be added dynamically as your events evolve.

        This dictionary can be helpful for a wide variety of use cases:

        - Numeric properties on events like `compute_time_ms` can later be inputs to our
          flexible query engine to determine usage.
        - Logging a region or cluster with each event can help you provide customers
          more granular visibility into their usage.

        We encourage logging this metadata with an eye towards future use cases to
        ensure full coverage for historical data. The datatype of the value in the
        properties dictionary is important for metric creation from an event source.
        Values that you wish to numerically aggregate should be of numeric type in the
        event.

        ## Determining event timestamp

        For cases where usage is being reported in real time as it is occurring,
        timestamp should correspond to the time that usage occurred.

        In cases where usage is reported in aggregate for a historical timeframe at a
        regular interval, we recommend setting the event `timestamp` to the midpoint of
        the interval. As an example, if you have an hourly reporter that sends data once
        an hour for the previous hour of usage, setting the `timestamp` to the half-hour
        mark will ensure that the usage is counted within the correct period.

        Note that other time-related fields (e.g. time elapsed) can be added to the
        properties dictionary as necessary.

        In cases where usage is reported in aggregate for a historical timeframe, the
        timestamp must be within the grace period set for your account. Events with
        `timestamp < current_time - grace_period` will not be accepted as a valid event,
        and will throw validation errors. Enforcing the grace period enables Orb to
        accurately map usage to the correct billing cycle and ensure that all usage is
        billed for in the corresponding billing period.

        ## Event validation

        Orb’s validation ensures that you recognize errors in your events as quickly as
        possible, and the API provides informative error messages to help you fix
        problems quickly.

        We validate the following:

        - Exactly one of `customer_id` and `external_customer_id` should be specified.
        - If specified, `customer_id` must identify a Customer resource within Orb. We
          do not support sending events for customers that have not been provisioned.
          Similarly, if specified, `external_customer_id` must be an identifier that is
          associated with an Orb Customer resource. Note: During our initial integration
          period, this enforcement will be temporarily turned into a warning to ensure
          smooth customer migration.
        - `timestamp` must conform to ISO 8601 and represent a timestamp at most 1 hour
          in the future. This timestamp should be sent in UTC timezone (no timezone
          offset).

        ## Idempotency and retry semantics

        Orb's idempotency guarantees allow you to implement safe retry logic in the
        event of network or machine failures, ensuring data fidelity. Each event in the
        request payload is associated with an idempotency key, and Orb guarantees that a
        single idempotency key will be successfully ingested at most once.

        - Successful responses return a 200 HTTP status code. The response contains
          information about previously processed events.
        - Requests that return a `4xx` HTTP status code indicate a payload error and
          contain at least one event with a validation failure. An event with a
          validation failure can be re-sent to the ingestion endpoint (after the payload
          is fixed) with the original idempotency key since that key is not marked as
          processed.
        - Requests that return a `5xx` HTTP status code indicate a server-side failure.
          These requests should be retried in their entirety.

        ## API usage and limits

        The ingestion API is designed made for real-time streaming ingestion and
        architected for high throughput. Even if events are later deemed unnecessary or
        filtered out, we encourage you to log them to Orb if they may be relevant to
        billing calculations in the future.

        To take advantage of the real-time features of the Orb platform and avoid any
        chance of dropped events by producers, we recommend reporting events to Orb
        frequently. Optionally, events can also be briefly aggregated at the source, as
        this API accepts an array of event bodies.

        Orb does not currently enforce a hard rate-limit for API usage or a maximum
        request payload size, but please give us a heads up if you’re changing either of
        these factors by an order of magnitude from initial setup.

        ## Testing in debug mode

        The ingestion API supports a debug mode, which returns additional verbose output
        to indicate which event idempotency keys were newly ingested or duplicates from
        previous requests. To enable this mode, mark `debug=true` as a query parameter.

        If `debug=true` is not specified, the response will only contain
        `validation_failed`. Orb will still honor the idempotency guarantees set
        [here](../guides/events-and-metrics/event-ingestion#event-volume-and-concurrency)
        in all cases.

        We strongly recommend that you only use debug mode as part of testing your
        initial Orb integration. Once you're ready to switch to production, disable
        debug mode to take advantage of improved performance and maximal throughput.

        #### Example: ingestion response with `debug=true`

        ```json
        {
          "debug": {
            "duplicate": [],
            "ingested": ["B7E83HDMfJPAunXW", "SJs5DQJ3TnwSqEZE", "8SivfDsNKwCeAXim"]
          },
          "validation_failed": []
        }
        ```

        #### Example: ingestion response with `debug=false`

        ```json
        {
          "validation_failed": []
        }
        ```

        Args:
          backfill_id: If this ingestion request is part of a backfill, this parameter ties the
              ingested events to the backfill

          debug: Flag to enable additional debug information in the endpoint response

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._post(
            "/ingest",
            body=maybe_transform({"events": events}, event_ingest_params.EventIngestParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
                query=maybe_transform(
                    {
                        "backfill_id": backfill_id,
                        "debug": debug,
                    },
                    event_ingest_params.EventIngestParams,
                ),
            ),
            cast_to=EventIngestResponse,
        )

    async def search(
        self,
        *,
        event_ids: List[str],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> EventSearchResponse:
        """
        This endpoint returns a filtered set of events for an account in a
        [paginated list format](../reference/pagination).

        Note that this is a `POST` endpoint rather than a `GET` endpoint because it
        employs a JSON body for search criteria rather than query parameters, allowing
        for a more flexible search syntax.

        Note that a search criteria _must_ be specified. Currently, Orb supports the
        following criteria:

        - `event_ids`: This is an explicit array of IDs to filter by. Note that an
          event's ID is the `idempotency_key` that was originally used for ingestion.

        By default, Orb does not return _deprecated_ events in this endpoint.

        By default, Orb will not throw a `404` if no events matched, Orb will return an
        empty array for `data` instead.

        Args:
          event_ids: This is an explicit array of IDs to filter by. Note that an event's ID is the
              idempotency_key that was originally used for ingestion. Values in this array
              will be treated case sensitively.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._post(
            "/events/search",
            body=maybe_transform({"event_ids": event_ids}, event_search_params.EventSearchParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=EventSearchResponse,
        )


class EventsWithRawResponse:
    def __init__(self, events: Events) -> None:
        self.backfills = BackfillsWithRawResponse(events.backfills)

        self.update = to_raw_response_wrapper(
            events.update,
        )
        self.deprecate = to_raw_response_wrapper(
            events.deprecate,
        )
        self.ingest = to_raw_response_wrapper(
            events.ingest,
        )
        self.search = to_raw_response_wrapper(
            events.search,
        )


class AsyncEventsWithRawResponse:
    def __init__(self, events: AsyncEvents) -> None:
        self.backfills = AsyncBackfillsWithRawResponse(events.backfills)

        self.update = async_to_raw_response_wrapper(
            events.update,
        )
        self.deprecate = async_to_raw_response_wrapper(
            events.deprecate,
        )
        self.ingest = async_to_raw_response_wrapper(
            events.ingest,
        )
        self.search = async_to_raw_response_wrapper(
            events.search,
        )



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Dict, Union, Optional
from datetime import date
from typing_extensions import Literal, Required, Annotated, TypedDict

from ...._utils import PropertyInfo

__all__ = [
    "LedgerCreateEntryParams",
    "AddIncrementCreditLedgerEntryRequestParams",
    "AddIncrementCreditLedgerEntryRequestParamsInvoiceSettings",
    "AddDecrementCreditLedgerEntryRequestParams",
    "AddExpirationChangeCreditLedgerEntryRequestParams",
    "AddVoidCreditLedgerEntryRequestParams",
    "AddAmendmentCreditLedgerEntryRequestParams",
]


class AddIncrementCreditLedgerEntryRequestParams(TypedDict, total=False):
    amount: Required[float]
    """The number of credits to effect.

    Note that this is required for increment, decrement, void, or undo operations.
    """

    entry_type: Required[Literal["increment"]]

    description: Optional[str]
    """Optional metadata that can be specified when adding ledger results via the API.

    For example, this can be used to note an increment refers to trial credits, or
    for noting corrections as a result of an incident, etc.
    """

    effective_date: Annotated[Union[str, date, None], PropertyInfo(format="iso8601")]
    """
    A future date (specified in YYYY-MM-DD format) that denotes when this credit
    balance should become available for use.
    """

    expiry_date: Annotated[Union[str, date, None], PropertyInfo(format="iso8601")]
    """
    A future date (specified in YYYY-MM-DD format) that denotes when this credit
    balance should expire.
    """

    invoice_settings: Optional[AddIncrementCreditLedgerEntryRequestParamsInvoiceSettings]
    """
    Passing `invoice_settings` automatically generates an invoice for the newly
    added credits. If `invoice_settings` is passed, you must specify
    per_unit_cost_basis, as the calculation of the invoice total is done on that
    basis.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    per_unit_cost_basis: Optional[str]
    """Can only be specified when entry_type=increment.

    How much, in USD, a customer paid for a single credit in this block
    """


class AddIncrementCreditLedgerEntryRequestParamsInvoiceSettings(TypedDict, total=False):
    auto_collection: Required[bool]
    """
    Whether the credits purchase invoice should auto collect with the customer's
    saved payment method.
    """

    net_terms: Required[int]
    """
    The net terms determines the difference between the invoice date and the issue
    date for the invoice. If you intend the invoice to be due on issue, set this
    to 0.
    """

    memo: Optional[str]
    """An optional memo to display on the invoice."""


class AddDecrementCreditLedgerEntryRequestParams(TypedDict, total=False):
    amount: Required[float]
    """The number of credits to effect.

    Note that this is required for increment, decrement, void, or undo operations.
    """

    entry_type: Required[Literal["decrement"]]

    description: Optional[str]
    """Optional metadata that can be specified when adding ledger results via the API.

    For example, this can be used to note an increment refers to trial credits, or
    for noting corrections as a result of an incident, etc.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """


class AddExpirationChangeCreditLedgerEntryRequestParams(TypedDict, total=False):
    entry_type: Required[Literal["expiration_change"]]

    expiry_date: Required[Annotated[Union[str, date, None], PropertyInfo(format="iso8601")]]
    """
    A future date (specified in YYYY-MM-DD format) that identifies the origination
    credit block to expire
    """

    target_expiry_date: Required[Annotated[Union[str, date], PropertyInfo(format="iso8601")]]
    """
    A future date (specified in YYYY-MM-DD format) used for expiration change,
    denoting when credits transferred (as part of a partial block expiration) should
    expire.
    """

    amount: Optional[float]
    """The number of credits to effect.

    Note that this is required for increment, decrement, void, or undo operations.
    """

    block_id: Optional[str]
    """
    The ID of the block affected by an expiration_change, used to differentiate
    between multiple blocks with the same `expiry_date`.
    """

    description: Optional[str]
    """Optional metadata that can be specified when adding ledger results via the API.

    For example, this can be used to note an increment refers to trial credits, or
    for noting corrections as a result of an incident, etc.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """


class AddVoidCreditLedgerEntryRequestParams(TypedDict, total=False):
    amount: Required[float]
    """The number of credits to effect.

    Note that this is required for increment, decrement, void, or undo operations.
    """

    block_id: Required[str]
    """The ID of the block to void."""

    entry_type: Required[Literal["void"]]

    description: Optional[str]
    """Optional metadata that can be specified when adding ledger results via the API.

    For example, this can be used to note an increment refers to trial credits, or
    for noting corrections as a result of an incident, etc.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    void_reason: Optional[Literal["refund"]]
    """Can only be specified when `entry_type=void`. The reason for the void."""


class AddAmendmentCreditLedgerEntryRequestParams(TypedDict, total=False):
    amount: Required[float]
    """The number of credits to effect.

    Note that this is required for increment, decrement or void operations.
    """

    block_id: Required[str]
    """The ID of the block to reverse a decrement from."""

    entry_type: Required[Literal["amendment"]]

    description: Optional[str]
    """Optional metadata that can be specified when adding ledger results via the API.

    For example, this can be used to note an increment refers to trial credits, or
    for noting corrections as a result of an incident, etc.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """


LedgerCreateEntryParams = Union[
    AddIncrementCreditLedgerEntryRequestParams,
    AddDecrementCreditLedgerEntryRequestParams,
    AddExpirationChangeCreditLedgerEntryRequestParams,
    AddVoidCreditLedgerEntryRequestParams,
    AddAmendmentCreditLedgerEntryRequestParams,
]



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from .ledger_list_params import LedgerListParams as LedgerListParams
from .ledger_list_response import LedgerListResponse as LedgerListResponse
from .ledger_create_entry_params import (
    LedgerCreateEntryParams as LedgerCreateEntryParams,
)
from .ledger_create_entry_response import (
    LedgerCreateEntryResponse as LedgerCreateEntryResponse,
)
from .ledger_list_by_external_id_params import (
    LedgerListByExternalIDParams as LedgerListByExternalIDParams,
)
from .ledger_list_by_external_id_response import (
    LedgerListByExternalIDResponse as LedgerListByExternalIDResponse,
)
from .ledger_create_entry_by_external_id_params import (
    LedgerCreateEntryByExternalIDParams as LedgerCreateEntryByExternalIDParams,
)
from .ledger_create_entry_by_external_id_response import (
    LedgerCreateEntryByExternalIDResponse as LedgerCreateEntryByExternalIDResponse,
)



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Union
from datetime import date
from typing_extensions import Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["SubscriptionTriggerPhaseParams"]


class SubscriptionTriggerPhaseParams(TypedDict, total=False):
    effective_date: Annotated[Union[str, date, None], PropertyInfo(format="iso8601")]
    """The date on which the phase change should take effect.

    If not provided, defaults to today in the customer's timezone.
    """



# File generated from our OpenAPI spec by Stainless.

from __future__ import annotations

from typing import Dict, List, Union, Optional
from typing_extensions import Literal, Required, TypedDict

__all__ = [
    "SubscriptionSchedulePlanChangeParams",
    "PriceOverride",
    "PriceOverrideOverrideUnitPrice",
    "PriceOverrideOverrideUnitPriceUnitConfig",
    "PriceOverrideOverrideUnitPriceDiscount",
    "PriceOverrideOverridePackagePrice",
    "PriceOverrideOverridePackagePricePackageConfig",
    "PriceOverrideOverridePackagePriceDiscount",
    "PriceOverrideOverrideMatrixPrice",
    "PriceOverrideOverrideMatrixPriceMatrixConfig",
    "PriceOverrideOverrideMatrixPriceMatrixConfigMatrixValue",
    "PriceOverrideOverrideMatrixPriceDiscount",
    "PriceOverrideOverrideTieredPrice",
    "PriceOverrideOverrideTieredPriceTieredConfig",
    "PriceOverrideOverrideTieredPriceTieredConfigTier",
    "PriceOverrideOverrideTieredPriceDiscount",
    "PriceOverrideOverrideTieredBpsPrice",
    "PriceOverrideOverrideTieredBpsPriceTieredBpsConfig",
    "PriceOverrideOverrideTieredBpsPriceTieredBpsConfigTier",
    "PriceOverrideOverrideTieredBpsPriceDiscount",
    "PriceOverrideOverrideBpsPrice",
    "PriceOverrideOverrideBpsPriceBpsConfig",
    "PriceOverrideOverrideBpsPriceDiscount",
    "PriceOverrideOverrideBulkBpsPrice",
    "PriceOverrideOverrideBulkBpsPriceBulkBpsConfig",
    "PriceOverrideOverrideBulkBpsPriceBulkBpsConfigTier",
    "PriceOverrideOverrideBulkBpsPriceDiscount",
    "PriceOverrideOverrideBulkPrice",
    "PriceOverrideOverrideBulkPriceBulkConfig",
    "PriceOverrideOverrideBulkPriceBulkConfigTier",
    "PriceOverrideOverrideBulkPriceDiscount",
    "PriceOverrideOverrideTestRatingFunctionPrice",
    "PriceOverrideOverrideTestRatingFunctionPriceDiscount",
    "PriceOverrideOverrideFivetranExamplePrice",
    "PriceOverrideOverrideFivetranExamplePriceDiscount",
    "PriceOverrideOverrideThresholdTotalAmountPrice",
    "PriceOverrideOverrideThresholdTotalAmountPriceDiscount",
    "PriceOverrideOverrideTieredPackagePrice",
    "PriceOverrideOverrideTieredPackagePriceDiscount",
    "PriceOverrideOverrideTieredWithMinimumPrice",
    "PriceOverrideOverrideTieredWithMinimumPriceDiscount",
    "PriceOverrideOverridePackageWithAllocationPrice",
    "PriceOverrideOverridePackageWithAllocationPriceDiscount",
]


class SubscriptionSchedulePlanChangeParams(TypedDict, total=False):
    change_option: Required[Literal["requested_date", "end_of_subscription_term", "immediate"]]

    align_billing_with_plan_change_date: Optional[bool]
    """[DEPRECATED] Use billing_cycle_alignment instead.

    Reset billing periods to be aligned with the plan change’s effective date.
    """

    billing_cycle_alignment: Optional[Literal["unchanged", "plan_change_date", "start_of_month"]]
    """
    Reset billing periods to be aligned with the plan change’s effective date or
    start of the month. Defaults to `unchanged` which keeps subscription's existing
    billing cycle alignment.
    """

    change_date: Optional[str]
    """The date that the plan change should take effect.

    This parameter can only be passed if the `change_option` is `requested_date`.
    """

    coupon_redemption_code: Optional[str]
    """Redemption code to be used for this subscription.

    If the coupon cannot be found by its redemption code, or cannot be redeemed, an
    error response will be returned and the plan change will not be scheduled.
    """

    credits_overage_rate: Optional[float]

    external_plan_id: Optional[str]
    """
    The external_plan_id of the plan that the given subscription should be switched
    to. Note that either this property or `plan_id` must be specified.
    """

    initial_phase_order: Optional[int]
    """The phase of the plan to start with"""

    invoicing_threshold: Optional[str]
    """
    When this subscription's accrued usage reaches this threshold, an invoice will
    be issued for the subscription. If not specified, invoices will only be issued
    at the end of the billing period.
    """

    per_credit_overage_amount: Optional[str]

    plan_id: Optional[str]
    """The plan that the given subscription should be switched to.

    Note that either this property or `external_plan_id` must be specified.
    """

    price_overrides: Optional[List[PriceOverride]]
    """Optionally provide a list of overrides for prices on the plan"""


class PriceOverrideOverrideUnitPriceUnitConfig(TypedDict, total=False):
    unit_amount: Required[str]
    """Rate per unit of usage"""

    scaling_factor: Optional[float]
    """Multiplier to scale rated quantity by"""


class PriceOverrideOverrideUnitPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideUnitPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["unit"]]

    unit_config: Required[PriceOverrideOverrideUnitPriceUnitConfig]

    discount: Optional[PriceOverrideOverrideUnitPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverridePackagePricePackageConfig(TypedDict, total=False):
    package_amount: Required[str]
    """A currency amount to rate usage by"""

    package_size: Optional[int]
    """An integer amount to represent package size.

    For example, 1000 here would divide usage by 1000 before multiplying by
    package_amount in rating
    """


class PriceOverrideOverridePackagePriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverridePackagePrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["package"]]

    package_config: Required[PriceOverrideOverridePackagePricePackageConfig]

    discount: Optional[PriceOverrideOverridePackagePriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideMatrixPriceMatrixConfigMatrixValue(TypedDict, total=False):
    dimension_values: Required[List[Optional[str]]]
    """One or two matrix keys to filter usage to this Matrix value by.

    For example, ["region", "tier"] could be used to filter cloud usage by a cloud
    region and an instance tier.
    """

    unit_amount: Required[str]
    """Unit price for the specified dimension_values"""

    scaling_factor: Optional[float]
    """Optional multiplier to scale rated quantities by"""


class PriceOverrideOverrideMatrixPriceMatrixConfig(TypedDict, total=False):
    default_unit_amount: Required[str]
    """Default per unit rate for any usage not bucketed into a specified matrix_value"""

    dimensions: Required[List[Optional[str]]]
    """One or two event property values to evaluate matrix groups by"""

    matrix_values: Required[List[PriceOverrideOverrideMatrixPriceMatrixConfigMatrixValue]]
    """Matrix values for specified matrix grouping keys"""

    scaling_factor: Optional[float]
    """
    Default optional multiplier to scale rated quantities that fall into the default
    bucket by
    """


class PriceOverrideOverrideMatrixPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideMatrixPrice(TypedDict, total=False):
    id: Required[str]

    matrix_config: Required[PriceOverrideOverrideMatrixPriceMatrixConfig]

    model_type: Required[Literal["matrix"]]

    discount: Optional[PriceOverrideOverrideMatrixPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTieredPriceTieredConfigTier(TypedDict, total=False):
    first_unit: Required[float]
    """Inclusive tier starting value"""

    unit_amount: Required[str]
    """Amount per unit"""

    last_unit: Optional[float]
    """Exclusive tier ending value. If null, this is treated as the last tier"""


class PriceOverrideOverrideTieredPriceTieredConfig(TypedDict, total=False):
    tiers: Required[List[PriceOverrideOverrideTieredPriceTieredConfigTier]]
    """Tiers for rating based on total usage quantities into the specified tier"""


class PriceOverrideOverrideTieredPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTieredPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["tiered"]]

    tiered_config: Required[PriceOverrideOverrideTieredPriceTieredConfig]

    discount: Optional[PriceOverrideOverrideTieredPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTieredBpsPriceTieredBpsConfigTier(TypedDict, total=False):
    bps: Required[float]
    """Per-event basis point rate"""

    minimum_amount: Required[str]
    """Inclusive tier starting value"""

    maximum_amount: Optional[str]
    """Exclusive tier ending value"""

    per_unit_maximum: Optional[str]
    """Per unit maximum to charge"""


class PriceOverrideOverrideTieredBpsPriceTieredBpsConfig(TypedDict, total=False):
    tiers: Required[List[PriceOverrideOverrideTieredBpsPriceTieredBpsConfigTier]]
    """
    Tiers for a Graduated BPS pricing model, where usage is bucketed into specified
    tiers
    """


class PriceOverrideOverrideTieredBpsPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTieredBpsPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["tiered_bps"]]

    tiered_bps_config: Required[PriceOverrideOverrideTieredBpsPriceTieredBpsConfig]

    discount: Optional[PriceOverrideOverrideTieredBpsPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideBpsPriceBpsConfig(TypedDict, total=False):
    bps: Required[float]
    """Basis point take rate per event"""

    per_unit_maximum: Optional[str]
    """Optional currency amount maximum to cap spend per event"""


class PriceOverrideOverrideBpsPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideBpsPrice(TypedDict, total=False):
    id: Required[str]

    bps_config: Required[PriceOverrideOverrideBpsPriceBpsConfig]

    model_type: Required[Literal["bps"]]

    discount: Optional[PriceOverrideOverrideBpsPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideBulkBpsPriceBulkBpsConfigTier(TypedDict, total=False):
    bps: Required[float]
    """Basis points to rate on"""

    maximum_amount: Optional[str]
    """Upper bound for tier"""

    per_unit_maximum: Optional[str]
    """The maximum amount to charge for any one event"""


class PriceOverrideOverrideBulkBpsPriceBulkBpsConfig(TypedDict, total=False):
    tiers: Required[List[PriceOverrideOverrideBulkBpsPriceBulkBpsConfigTier]]
    """
    Tiers for a bulk BPS pricing model where all usage is aggregated to a single
    tier based on total volume
    """


class PriceOverrideOverrideBulkBpsPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideBulkBpsPrice(TypedDict, total=False):
    id: Required[str]

    bulk_bps_config: Required[PriceOverrideOverrideBulkBpsPriceBulkBpsConfig]

    model_type: Required[Literal["bulk_bps"]]

    discount: Optional[PriceOverrideOverrideBulkBpsPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideBulkPriceBulkConfigTier(TypedDict, total=False):
    unit_amount: Required[str]
    """Amount per unit"""

    maximum_units: Optional[float]
    """Upper bound for this tier"""


class PriceOverrideOverrideBulkPriceBulkConfig(TypedDict, total=False):
    tiers: Required[List[PriceOverrideOverrideBulkPriceBulkConfigTier]]
    """Bulk tiers for rating based on total usage volume"""


class PriceOverrideOverrideBulkPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideBulkPrice(TypedDict, total=False):
    id: Required[str]

    bulk_config: Required[PriceOverrideOverrideBulkPriceBulkConfig]

    model_type: Required[Literal["bulk"]]

    discount: Optional[PriceOverrideOverrideBulkPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTestRatingFunctionPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTestRatingFunctionPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["test_rating_function"]]

    test_rating_function_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverrideTestRatingFunctionPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideFivetranExamplePriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideFivetranExamplePrice(TypedDict, total=False):
    id: Required[str]

    fivetran_example_config: Required[Dict[str, object]]

    model_type: Required[Literal["fivetran_example"]]

    discount: Optional[PriceOverrideOverrideFivetranExamplePriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideThresholdTotalAmountPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideThresholdTotalAmountPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["threshold_total_amount"]]

    threshold_total_amount_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverrideThresholdTotalAmountPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTieredPackagePriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTieredPackagePrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["tiered_package"]]

    tiered_package_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverrideTieredPackagePriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverrideTieredWithMinimumPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverrideTieredWithMinimumPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["tiered_with_minimum"]]

    tiered_with_minimum_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverrideTieredWithMinimumPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


class PriceOverrideOverridePackageWithAllocationPriceDiscount(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "trial", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    applies_to_price_ids: Optional[List[str]]
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    trial_amount_discount: Optional[str]
    """Only available if discount_type is `trial`"""

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """


class PriceOverrideOverridePackageWithAllocationPrice(TypedDict, total=False):
    id: Required[str]

    model_type: Required[Literal["package_with_allocation"]]

    package_with_allocation_config: Required[Dict[str, object]]

    discount: Optional[PriceOverrideOverridePackageWithAllocationPriceDiscount]
    """The subscription's override discount for the plan."""

    fixed_price_quantity: Optional[float]
    """The starting quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """The subscription's override maximum amount for the plan."""

    minimum_amount: Optional[str]
    """The subscription's override minimum amount for the plan."""


PriceOverride = Union[
    PriceOverrideOverrideUnitPrice,
    PriceOverrideOverridePackagePrice,
    PriceOverrideOverrideMatrixPrice,
    PriceOverrideOverrideTieredPrice,
    PriceOverrideOverrideTieredBpsPrice,
    PriceOverrideOverrideBpsPrice,
    PriceOverrideOverrideBulkBpsPrice,
    PriceOverrideOverrideBulkPrice,
    PriceOverrideOverrideTestRatingFunctionPrice,
    PriceOverrideOverrideFivetranExamplePrice,
    PriceOverrideOverrideThresholdTotalAmountPrice,
    PriceOverrideOverrideTieredPackagePrice,
    PriceOverrideOverrideTieredWithMinimumPrice,
    PriceOverrideOverridePackageWithAllocationPrice,
]
